#include <windows.h>
#include <stdio.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int filterRunProc(const FilterActivation *fa, const FilterFunctions *ff);
int filterStartProc(FilterActivation *fa, const FilterFunctions *ff);
int filterEndProc(FilterActivation *fa, const FilterFunctions *ff);
long filterParamProc(FilterActivation *fa, const FilterFunctions *ff);
int filterConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void filterStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void filterScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool filterFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);

///////////////////////////////////////////////////////////////////////////

typedef struct MyFilterData {
    bool fInterpolate;
} MyFilterData;

ScriptFunctionDef filter_func_defs[]={
    { (ScriptFunctionPtr)filterScriptConfig, "Config", "0i" },
    { NULL },
};

CScriptObject filter_obj={
    NULL, filter_func_defs
};

struct FilterDefinition filterDef = {

    NULL, NULL, NULL,       // next, prev, module
    "4:1:1 Helper",             // name
    "Converts 4:1:1 to 4:4:4",
                            // desc
    "Rick Tillery",               // maker
    NULL,                   // private_data
    sizeof(MyFilterData),   // inst_data_size

    NULL,                   // initProc
    NULL,                   // deinitProc
    filterRunProc,        // runProc
    filterParamProc,      // paramProc
    filterConfigProc,     // configProc
    filterStringProc,     // stringProc
    filterStartProc,      // startProc
    filterEndProc,        // endProc

    &filter_obj,          // script_obj
    filterFssProc,        // fssProc

};

///////////////////////////////////////////////////////////////////////////

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_filter;

///////////////////////////////////////////////////////////////////////////

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
    if (!(fd_filter = ff->addFilter(fm, &filterDef, sizeof(FilterDefinition))))
        return 1;

  vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
  vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;

    return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
    ff->removeFilter(fd_filter);
}

///////////////////////////////////////////////////////////////////////////

static bool g_MMXenabled;

int filterStartProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    g_MMXenabled = ff->isMMXEnabled();

    return 0;
}
/*
void __declspec(naked) doscan_MMX(Pixel32 *dst, Pixel32 *src, int w, long frac, long bias, int fDouble) {
    static const __int64 Rmask = 0x0000FFFF00000000i64;

    __asm {
        push        ebp
        push        edi
        push        esi
        push        ebx

        mov         eax,[esp+4+16]
        mov         edx,[esp+8+16]

        mov         ecx,[esp+12+16]
        neg         ecx
        shl         ecx,2
        sub         eax,ecx
        sub         edx,ecx

        movq        mm6,Rmask
        movd        mm4,[esp+20+16]
        psllq       mm4,16
        movd        mm5,[esp+16+16]
        punpcklwd   mm5,mm5
        pxor        mm7,mm7

        mov         ebx,dword ptr [esp+24+16]
        or          ebx,ebx
        jz          xloop1

        sub         eax,ecx

xloop2:
        movd        mm0,[edx+ecx]       ;mm0 = pixel
        movq        mm1,mm6             ;mm1 = R mask
        punpcklbw   mm0,mm7             ;unpack pixel to words
        pand        mm1,mm0             ;mm1 = red component
        pmulhw      mm0,mm5             ;scale green and blue
        paddw       mm1,mm4             ;add green bias
        paddw       mm0,mm1             ;add scaled green/blue
        packuswb    mm0,mm0             ;repack pixel to bytes
        movq        [eax+ecx*2],mm0     ;write 2 pixels
        add         ecx,4
        jne         xloop2
    jmp      short xit

xloop1:
        movd        mm0,[edx+ecx]       ;mm0 = pixel
        movq        mm1,mm6             ;mm1 = R mask
        punpcklbw   mm0,mm7             ;unpack pixel to words
        pand        mm1,mm0             ;mm1 = red component
        paddw       mm0,mm0             ;double g/b channels beforehand
        pmulhw      mm0,mm5             ;scale green and blue
        paddw       mm1,mm4             ;add green bias
        paddw       mm0,mm1             ;add scaled green/blue
        packuswb    mm0,mm0             ;repack pixel to bytes
        movd        [eax+ecx],mm0       ;write pixel
        add         ecx,4
        jne         xloop1

xit:
        pop         ebx
        pop         esi
        pop         edi
        pop         ebp
        ret
    }
}
*/
int filterRunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
  MyFilterData* mfd = (MyFilterData*)fa->filter_data;
  PixDim w, h;
  Pixel32* src;
  Pixel32* dst;

  src = (Pixel32*)fa->src.data;
  dst = (Pixel32*)fa->dst.data;

  h = fa->src.h;
  do
  {
    w = fa->src.w;
    do
    {
      Pixel32 pxl;
      int red, green, blue;
      int Y[5], RminusY[5], GminusY[5], BminusY[5];
      int new_red, new_green, new_blue;
      int deltaY, deltaRY, deltaGY, deltaBY;
      int temp, i;

      for(i = 0; i < 5; i++)
      {
        if(w > i)
        {
          pxl = *src;

          red = (int)((pxl >> 16) & 0xFF);
          green = (int)((WORD)pxl >> 8);
          blue = (int)(pxl & 0xFF);

          //  y = 0.299*r + 0.587*g + 0.114*b
          Y[i] = (19595 * red) + (38470 * green) + (7471 * blue);  // y * 65536

          if(!(i & 3))
          {
            RminusY[i] = (red * 65536) - Y[i];   // (r - y) * 65536

            GminusY[i] = (green * 65536) - Y[i]; // (g - y) * 65536

            BminusY[i] = (blue * 65536) - Y[i];  // (b - y) * 65536
          }

          src++;
        }
        else
        {
          Y[i] = Y[i - 1];
          RminusY[i] = RminusY[i - 1];
          GminusY[i] = GminusY[i - 1];
          BminusY[i] = BminusY[i - 1];

          if(i == 4)
            src++;
        }
      }
      --src;

      deltaY = (Y[4] - Y[0]) / 256;               // delta y * 256
      deltaRY = (RminusY[4] - RminusY[0]) / 256;  // delta (r - y) * 256
      deltaGY = (GminusY[4] - GminusY[0]) / 256;  // delta (g - y) * 256
      deltaBY = (BminusY[4] - BminusY[0]) / 256;  // delta (b - y) * 256

      if(deltaY && !mfd->fInterpolate)
      {
        if(deltaY > 0)
          if((Y[1] > Y[0]) && (Y[1] < Y[4]))
            temp = (Y[1] - Y[0]) / 256;                 // sub delta y * 256
          else
            temp = 0;
        else
          if((Y[1] < Y[0]) && (Y[1] > Y[4]))
            temp = (Y[1] - Y[0]) / 256;                 // sub delta y * 256
          else
            temp = 0;

        RminusY[1] = RminusY[0] + (((temp * deltaRY) / deltaY) * 256);
        GminusY[1] = GminusY[0] + (((temp * deltaGY) / deltaY) * 256);
        BminusY[1] = BminusY[0] + (((temp * deltaBY) / deltaY) * 256);

        if(deltaY > 0)
          if((Y[2] > Y[0]) && (Y[2] < Y[4]))
            temp = (Y[2] - Y[0]) / 256;                 // sub delta y * 256
          else
            temp = 0;
        else
          if((Y[2] < Y[0]) && (Y[2] > Y[4]))
            temp = (Y[2] - Y[0]) / 256;                 // sub delta y * 256
          else
            temp = 0;

        RminusY[2] = RminusY[0] + (((temp * deltaRY) / deltaY) * 256);
        GminusY[2] = GminusY[0] + (((temp * deltaGY) / deltaY) * 256);
        BminusY[2] = BminusY[0] + (((temp * deltaBY) / deltaY) * 256);

        if(deltaY > 0)
          if((Y[3] > Y[0]) && (Y[3] < Y[4]))
            temp = (Y[3] - Y[0]) / 256;                 // sub delta y * 256
          else
            temp = 0;
        else
          if((Y[3] < Y[0]) && (Y[3] > Y[4]))
            temp = (Y[3] - Y[0]) / 256;                 // sub delta y * 256
          else
            temp = 0;

        RminusY[3] = RminusY[0] + (((temp * deltaRY) / deltaY) * 256);
        GminusY[3] = GminusY[0] + (((temp * deltaGY) / deltaY) * 256);
        BminusY[3] = BminusY[0] + (((temp * deltaBY) / deltaY) * 256);
      }
      else
      {
        int deltaRYdiv4 = deltaRY * 64;
        RminusY[1] = RminusY[0] + deltaRYdiv4;
        RminusY[2] = RminusY[1] + deltaRYdiv4;
        RminusY[3] = RminusY[2] + deltaRYdiv4;

        int deltaGYdiv4 = deltaGY * 64;
        GminusY[1] = GminusY[0] + deltaGYdiv4;
        GminusY[2] = GminusY[1] + deltaGYdiv4;
        GminusY[3] = GminusY[2] + deltaGYdiv4;

        int deltaBYdiv4 = deltaBY * 64;
        BminusY[1] = BminusY[0] + deltaBYdiv4;
        BminusY[2] = BminusY[1] + deltaBYdiv4;
        BminusY[3] = BminusY[2] + deltaBYdiv4;
      }

      for(i = 0; i < 4; i++)
      {
        new_red = (Y[i] + RminusY[i]) / 65536;
        if(new_red < 0)
          new_red = 0;
        if(new_red > 255)
          new_red = 255;

        new_green = (Y[i] + GminusY[i]) / 65536;
        if(new_green < 0)
          new_green = 0;
        if(new_green > 255)
          new_green = 255;

        new_blue = (Y[i] + BminusY[i]) / 65536;
        if(new_blue < 0)
          new_blue = 0;
        if(new_blue > 255)
          new_blue = 255;

        pxl = (new_red  << 16) |
              (new_green << 8) |
              new_blue;

        *dst = pxl;
        dst++;
      }
      if(w >= 4)
        w -= 4;
      else
        w = 0;
    }while(w);

    src = (Pixel32 *)((char *)src + fa->src.modulo);
    dst = (Pixel32 *)((char *)dst + fa->dst.modulo);

/*
    if(g_MMXenabled)
    {
      doscan_MMX(dst, src, w,
                 mfd->fThird ? 0x2AAA : 0x4000,
                 mfd->fThird ? 0x55 : 0x80,
                 mfd->fExpand);

      src = (Pixel32*)((char*)src + fa->src.pitch);
      dst = (Pixel32*)((char*)dst + fa->dst.pitch);
    }
    else
    {
      do
      {
        old_pixel = *src++;
        *dst++ = new_pixel;
      }while(--w);
      src = (Pixel32 *)((char *)src + fa->src.modulo);
      dst = (Pixel32 *)((char *)dst + fa->dst.modulo);
    }

*/
  }while(--h);

  if(g_MMXenabled)
    __asm emms

  return(0);
}


int filterEndProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    return 0;
}

long filterParamProc(FilterActivation *fa, const FilterFunctions *ff) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

  fa->dst.pitch = (fa->dst.w*4 + 7) & -8;

    return FILTERPARAM_SWAP_BUFFERS;
}


BOOL CALLBACK filterConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam) {
    MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);

    switch(msg) {
        case WM_INITDIALOG:
            SetWindowLong(hdlg, DWL_USER, lParam);
            mfd = (MyFilterData *)lParam;

            CheckDlgButton(hdlg, IDC_LUMAFOLLOW, mfd->fInterpolate ? BST_UNCHECKED : BST_CHECKED);
            CheckDlgButton(hdlg, IDC_INTERPOLATE, mfd->fInterpolate ? BST_CHECKED : BST_UNCHECKED);

            return TRUE;

        case WM_COMMAND:
            switch(LOWORD(wParam))
            {
            case IDOK:
              mfd->fInterpolate = !!IsDlgButtonChecked(hdlg, IDC_INTERPOLATE);
              EndDialog(hdlg, 0);
              return TRUE;

            case IDCANCEL:
              EndDialog(hdlg, 1);
              return FALSE;

            case IDC_LUMAFOLLOW:
              mfd->fInterpolate = !IsDlgButtonChecked(hdlg, IDC_LUMAFOLLOW);
              CheckDlgButton(hdlg, IDC_INTERPOLATE, mfd->fInterpolate ? BST_CHECKED : BST_UNCHECKED);
              return(TRUE);

            case IDC_INTERPOLATE:
              mfd->fInterpolate = !!IsDlgButtonChecked(hdlg, IDC_INTERPOLATE);
              CheckDlgButton(hdlg, IDC_LUMAFOLLOW, mfd->fInterpolate ? BST_UNCHECKED : BST_CHECKED);
              return(TRUE);
            }
            break;
    }

    return FALSE;
}

int filterConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd) {
    return DialogBoxParam(fa->filter->module->hInstModule,
            MAKEINTRESOURCE(IDD_FILTER_DIALOG), hwnd,
            filterConfigDlgProc, (LPARAM)fa->filter_data);
}

void filterStringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str) {
    const char *modes[2]={
        " (Luma Follow)",
        " (Interpolate)",
    };
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    strcpy(str, modes[mfd->fInterpolate]);
}

void filterScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
    FilterActivation *fa = (FilterActivation *)lpVoid;
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    mfd->fInterpolate    = !!argv[0].asInt();
}

bool filterFssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
    MyFilterData *mfd = (MyFilterData *)fa->filter_data;

    _snprintf(buf, buflen, "Config(%d)",
        mfd->fInterpolate);

    return true;
}

